# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import urllib

import netaddr
import openstack
from openstack import exceptions as os_exc
from oslo_config import cfg
import tenacity

from ironic_inspector.common.i18n import _
from ironic_inspector.common import keystone
from ironic_inspector import utils

CONF = cfg.CONF
LOG = utils.getProcessingLogger(__name__)

# See https://docs.openstack.org/ironic/latest/contributor/states.html
VALID_STATES = frozenset(['enroll', 'manageable', 'inspecting', 'inspect wait',
                          'inspect failed'])

# States where an instance is deployed and an admin may be doing something.
VALID_ACTIVE_STATES = frozenset(['active', 'rescue'])

_IRONIC_SESSION = None


class NotFound(utils.Error):
    """Node not found in Ironic."""

    def __init__(self, node_ident, code=404, *args, **kwargs):
        msg = _('Node %s was not found in Ironic') % node_ident
        super(NotFound, self).__init__(msg, code, *args, **kwargs)


def _get_ironic_session():
    global _IRONIC_SESSION

    if not _IRONIC_SESSION:
        _IRONIC_SESSION = keystone.get_session('ironic')
    return _IRONIC_SESSION


def get_client(token=None):
    """Get an ironic client connection."""
    session = _get_ironic_session()

    try:
        return openstack.connection.Connection(
            session=session, oslo_conf=CONF).baremetal
    except Exception as exc:
        LOG.error('Failed to establish a connection with ironic, '
                  'reason: %s', exc)
        raise


def reset_ironic_session():
    """Reset the global session variable.

    Mostly useful for unit tests.
    """
    global _IRONIC_SESSION
    _IRONIC_SESSION = None


def get_ipmi_address(node):
    """Get the BMC address defined in node.driver_info dictionary

    Possible names of BMC address value examined in order of list
    ['ipmi_address'] + CONF.ipmi_address_fields. The value could
    be an IP address or a hostname. DNS lookup performed for the
    first non empty value.

    The first valid BMC address value returned along with
    it's v4 and v6 IP addresses.

    :param node: Node object with defined driver_info dictionary
    :return: tuple (ipmi_address, ipv4_address, ipv6_address)
    """
    none_address = None, None, None
    ipmi_fields = ['ipmi_address'] + CONF.ipmi_address_fields
    # NOTE(sambetts): IPMI Address is useless to us if bridging is enabled so
    # just ignore it and return None
    if node.driver_info.get("ipmi_bridging", "no") != "no":
        return none_address
    for name in ipmi_fields:
        value = node.driver_info.get(name)
        if not value:
            continue

        ipv4 = None
        ipv6 = None
        if '//' in value:
            url = urllib.parse.urlparse(value)
            value = url.hostname
        try:
            addrinfo = socket.getaddrinfo(value, None, 0, 0, socket.SOL_TCP)
            for family, socket_type, proto, canon_name, sockaddr in addrinfo:
                ip = sockaddr[0]
                if netaddr.IPAddress(ip).is_loopback():
                    LOG.warning('Ignoring loopback BMC address %s', ip,
                                node_info=node)
                elif family == socket.AF_INET:
                    ipv4 = ip
                elif family == socket.AF_INET6:
                    ipv6 = ip
        except socket.gaierror:
            msg = _('Failed to resolve the hostname (%(value)s)'
                    ' for node %(uuid)s')
            raise utils.Error(msg % {'value': value,
                                     'uuid': node.id},
                              node_info=node)

        return (value, ipv4, ipv6) if ipv4 or ipv6 else none_address
    return none_address


def check_provision_state(node):
    """Sanity checks the provision state of the node.

    :param node: An API client returned node object describing
                 the baremetal node according to ironic's node
                 data model.
    :returns: None if no action is to be taken, True if the
              power node state should not be modified.
    :raises: Error on an invalid state being detected.
    """
    state = node.provision_state.lower()
    if state not in VALID_STATES:
        if (CONF.processing.permit_active_introspection
                and state in VALID_ACTIVE_STATES):
            # Hey, we can leave the power on! Lets return
            # True to let the caller know.
            return True

        msg = _('Invalid provision state for introspection: '
                '"%(state)s", valid states are "%(valid)s"')
        raise utils.Error(msg % {'state': state,
                                 'valid': list(VALID_STATES)},
                          node_info=node)


def capabilities_to_dict(caps):
    """Convert the Node's capabilities into a dictionary."""
    if not caps:
        return {}
    return dict([key.split(':', 1) for key in caps.split(',')])


def dict_to_capabilities(caps_dict):
    """Convert a dictionary into a string with the capabilities syntax."""
    return ','.join(["%s:%s" % (key, value)
                     for key, value in caps_dict.items()
                     if value is not None])


def get_node(node_id, ironic=None, **kwargs):
    """Get a node from Ironic.

    :param node_id: node UUID or name.
    :param ironic: ironic client instance.
    :param kwargs: arguments to pass to Ironic client.
    :raises: Error on failure
    """
    ironic = ironic if ironic is not None else get_client()

    try:
        node = ironic.get_node(node_id, **kwargs)
    except os_exc.ResourceNotFound:
        raise NotFound(node_id)
    except os_exc.BadRequestException as exc:
        raise utils.Error(_("Cannot get node %(node)s: %(exc)s") %
                          {'node': node_id, 'exc': exc})
    return node


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(os_exc.SDKException),
    stop=tenacity.stop_after_attempt(5),
    wait=tenacity.wait_fixed(1),
    reraise=True)
def call_with_retries(func, *args, **kwargs):
    """Call an ironic client function retrying all errors.

    If an ironic client exception is raised, try calling the func again,
    at most 5 times, waiting 1 sec between each call. If on the 5th attempt
    the func raises again, the exception is propagated to the caller.
    """
    return func(*args, **kwargs)


def lookup_node_by_macs(macs, introspection_data=None,
                        ironic=None, fail=False):
    """Find a node by its MACs."""
    if ironic is None:
        ironic = get_client()

    nodes = set()
    for mac in macs:
        ports = ironic.ports(address=mac, fields=["uuid", "node_uuid"])
        ports = list(ports)
        if not ports:
            continue
        elif fail:
            raise utils.Error(
                _('Port %(mac)s already exists, uuid: %(uuid)s') %
                {'mac': mac, 'uuid': ports[0].id}, data=introspection_data)
        else:
            nodes.update(p.node_id for p in ports)

    if len(nodes) > 1:
        raise utils.Error(_('MAC addresses %(macs)s correspond to more than '
                            'one node: %(nodes)s') %
                          {'macs': ', '.join(macs),
                           'nodes': ', '.join(nodes)},
                          data=introspection_data)

    elif nodes:
        return nodes.pop()


def lookup_node_by_bmc_addresses(addresses, introspection_data=None,
                                 ironic=None, fail=False):
    """Find a node by its BMC address."""
    if ironic is None:
        ironic = get_client()

    # FIXME(aarefiev): it's not effective to fetch all nodes, and may
    #                  impact on performance on big clusters
    # TODO(TheJulia): We should likely first loop through nodes being
    #                 inspected, i.e. inspect wait, and then fallback
    #                 to the rest of the physical nodes so we limit
    #                 overall-impact of the operation.
    nodes = ironic.nodes(fields=('uuid', 'driver_info'), limit=None)
    found = set()
    for node in nodes:
        bmc_address, bmc_ipv4, bmc_ipv6 = get_ipmi_address(node)
        for addr in addresses:
            if addr not in (bmc_ipv4, bmc_ipv6):
                continue
            elif fail:
                raise utils.Error(
                    _('Node %(uuid)s already has BMC address %(addr)s') %
                    {'addr': addr, 'uuid': node.id},
                    data=introspection_data)
            else:
                found.add(node.id)

    if len(found) > 1:
        raise utils.Error(_('BMC addresses %(addr)s correspond to more than '
                            'one node: %(nodes)s') %
                          {'addr': ', '.join(addresses),
                           'nodes': ', '.join(found)},
                          data=introspection_data)
    elif found:
        return found.pop()


def lookup_node(macs=None, bmc_addresses=None, introspection_data=None,
                ironic=None):
    """Lookup a node in the ironic database."""
    node = node2 = None

    if macs:
        node = lookup_node_by_macs(macs, ironic=ironic)
    if bmc_addresses:
        node2 = lookup_node_by_bmc_addresses(bmc_addresses, ironic=ironic)

    if node and node2 and node != node2:
        raise utils.Error(_('MAC addresses %(mac)s and BMC addresses %(addr)s '
                            'correspond to different nodes: %(node1)s and '
                            '%(node2)s') %
                          {'mac': ', '.join(macs),
                           'addr': ', '.join(bmc_addresses),
                           'node1': node, 'node2': node2})

    return node or node2



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from openstack import exceptions
from oslo_config import cfg

from ironic_inspector import node_cache
from ironic_inspector.plugins import physnet_cidr_map
from ironic_inspector.test import base as test_base
from ironic_inspector import utils


class TestPhysnetCidrMapHook(test_base.NodeTest):
    hook = physnet_cidr_map.PhysnetCidrMapHook()

    def setUp(self):
        super(TestPhysnetCidrMapHook, self).setUp()
        self.data = {
            'inventory': {
                'interfaces': [{
                    'name': 'em1',
                    'mac_address': '11:11:11:11:11:11',
                    'ipv4_address': '1.1.1.1',
                }],
                'cpu': 1,
                'disks': 1,
                'memory': 1
            },
            'all_interfaces': {
                'em1': {},
            }
        }

        ports = [mock.Mock(spec=['address', 'id', 'physical_network'],
                           address=a) for a in ('11:11:11:11:11:11',)]
        self.node_info = node_cache.NodeInfo(uuid=self.uuid, started_at=0,
                                             node=self.node, ports=ports)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_expected_data(self, mock_patch):
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        patches = [{'path': '/physical_network',
                    'value': 'physnet_a',
                    'op': 'add'}]
        self.hook.before_update(self.data, self.node_info)
        self.assertCalledWithPatch(patches, mock_patch)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_no_matching_mapping_config(self, mock_patch):
        cfg.CONF.set_override('cidr_map', '2.2.2.0/24:physnet_b',
                              group='port_physnet')
        self.hook.before_update(self.data, self.node_info)
        self.assertFalse(mock_patch.called)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_expected_data_ipv6_lowercase(self, mock_patch):
        self.data['inventory']['interfaces'][0].pop('ipv4_address')
        self.data['inventory']['interfaces'][0]['ipv6_address'] = '2001:db8::1'
        cfg.CONF.set_override('cidr_map', '2001:db8::/64:physnet_b',
                              group='port_physnet')
        patches = [{'path': '/physical_network',
                    'value': 'physnet_b',
                    'op': 'add'}]
        self.hook.before_update(self.data, self.node_info)
        self.assertCalledWithPatch(patches, mock_patch)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_expected_data_ipv6_uppercase(self, mock_patch):
        self.data['inventory']['interfaces'][0].pop('ipv4_address')
        self.data['inventory']['interfaces'][0]['ipv6_address'] = '2001:db8::1'
        cfg.CONF.set_override('cidr_map', '2001:DB8::/64:physnet_b',
                              group='port_physnet')
        patches = [{'path': '/physical_network',
                    'value': 'physnet_b',
                    'op': 'add'}]
        self.hook.before_update(self.data, self.node_info)
        self.assertCalledWithPatch(patches, mock_patch)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_no_mapping_in_config(self, mock_patch):
        self.hook.before_update(self.data, self.node_info)
        self.assertFalse(mock_patch.called)

    def test_no_inventory(self):
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        del self.data['inventory']
        self.assertRaises(utils.Error, self.hook.before_update,
                          self.data, self.node_info)

    @mock.patch('ironic_inspector.plugins.base_physnet.LOG.debug',
                autospec=True)
    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_interface_not_in_ironic(self, mock_patch, mock_log):
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        self.node_info._ports = {}
        self.hook.before_update(self.data, self.node_info)
        self.assertTrue(mock_log.called)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_no_overwrite(self, mock_patch):
        ports = [mock.Mock(spec=['address', 'id', 'physical_network'],
                           address=a, physical_network='foo')
                 for a in ('11:11:11:11:11:11',)]
        node_info = node_cache.NodeInfo(uuid=self.uuid, started_at=0,
                                        node=self.node, ports=ports)
        cfg.CONF.set_override('overwrite_existing', False, group='processing')
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        self.hook.before_update(self.data, node_info)
        self.assertFalse(mock_patch.called)

    @mock.patch('ironic_inspector.plugins.base_physnet.LOG.warning',
                autospec=True)
    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_patch_port_exception(self, mock_patch, mock_log):
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        mock_patch.side_effect = exceptions.BadRequestException('invalid data')
        self.hook.before_update(self.data, self.node_info)
        log_msg = "Failed to update port %(uuid)s: %(error)s"
        mock_log.assert_called_with(log_msg, mock.ANY, node_info=mock.ANY)

    @mock.patch.object(node_cache.NodeInfo, 'patch_port', autospec=True)
    def test_no_ip_address_on_interface(self, mock_patch):
        cfg.CONF.set_override('cidr_map', '1.1.1.0/24:physnet_a',
                              group='port_physnet')
        data = {
            'inventory': {
                'interfaces': [{
                    'name': 'em1',
                    'mac_address': '11:11:11:11:11:11',
                }],
                'cpu': 1,
                'disks': 1,
                'memory': 1
            },
            'all_interfaces': {
                'em1': {},
            }
        }
        self.hook.before_update(data, self.node_info)
        self.assertFalse(mock_patch.called)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from ironic_inspector import db
from ironic_inspector.test import base as test_base


class TestDB(test_base.NodeTest):
    @mock.patch.object(db, 'get_reader_session', autospec=True)
    def test_model_query(self, mock_reader):
        mock_session = mock_reader.return_value
        fake_query = mock_session.query.return_value

        query = db.model_query('db.Node')

        mock_reader.assert_called_once_with()
        mock_session.query.assert_called_once_with('db.Node')
        self.assertEqual(fake_query, query)

    @mock.patch.object(db, 'get_writer_session', autospec=True)
    def test_ensure_transaction_new_session(self, mock_writer):
        mock_session = mock_writer.return_value

        with db.ensure_transaction() as session:
            mock_writer.assert_called_once_with()
            mock_session.begin.assert_called_once_with(subtransactions=True)
            self.assertEqual(mock_session, session)

    @mock.patch.object(db, 'get_writer_session', autospec=True)
    def test_ensure_transaction_session(self, mock_writer):
        mock_session = mock.MagicMock()

        with db.ensure_transaction(session=mock_session) as session:
            self.assertFalse(mock_writer.called)
            mock_session.begin.assert_called_once_with(subtransactions=True)
            self.assertEqual(mock_session, session)

    @mock.patch.object(db.enginefacade, 'transaction_context', autospec=True)
    def test__create_context_manager(self, mock_cnxt):
        mock_ctx_mgr = mock_cnxt.return_value

        ctx_mgr = db._create_context_manager()

        mock_ctx_mgr.configure.assert_called_once_with(sqlite_fk=False)
        self.assertEqual(mock_ctx_mgr, ctx_mgr)

    @mock.patch.object(db, 'get_context_manager', autospec=True)
    def test_get_reader_session(self, mock_cnxt_mgr):
        mock_cnxt = mock_cnxt_mgr.return_value
        mock_sess_maker = mock_cnxt.reader.get_sessionmaker.return_value

        session = db.get_reader_session()

        mock_sess_maker.assert_called_once_with()
        self.assertEqual(mock_sess_maker.return_value, session)

    @mock.patch.object(db, 'get_context_manager', autospec=True)
    def test_get_writer_session(self, mock_cnxt_mgr):
        mock_cnxt = mock_cnxt_mgr.return_value
        mock_sess_maker = mock_cnxt.writer.get_sessionmaker.return_value

        session = db.get_writer_session()

        mock_sess_maker.assert_called_once_with()
        self.assertEqual(mock_sess_maker.return_value, session)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic Rest Api tools."""

import functools

import flask
from oslo_config import cfg
from oslo_utils import uuidutils

from ironic_inspector.common.i18n import _
from ironic_inspector import utils

CONF = cfg.CONF


def raises_coercion_exceptions(fn):
    """Convert coercion function exceptions to utils.Error.

    :raises: utils.Error when the coercion function raises an
             AssertionError or a ValueError
    """
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            ret = fn(*args, **kwargs)
        except (AssertionError, ValueError) as exc:
            raise utils.Error(_('Bad request: %s') % exc, code=400)
        return ret
    return inner


def request_field(field_name):
    """Decorate a function that coerces the specified field.

    :param field_name: name of the field to fetch
    :returns: a decorator
    """
    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            default = kwargs.pop('default', None)
            field = flask.request.args.get(field_name, default=default)
            if field == default:
                # field not found or the same as the default, just return
                return default
            return fn(field, *args, **kwargs)
        return inner
    return outer


@request_field('marker')
@raises_coercion_exceptions
def marker_field(value):
    """Fetch the pagination marker field from flask.request.args.

    :returns: an uuid
    """
    assert uuidutils.is_uuid_like(value), _('Marker not UUID-like')
    return value


@request_field('limit')
@raises_coercion_exceptions
def limit_field(value):
    """Fetch the pagination limit field from flask.request.args.

    :returns: the limit
    """
    # limit of zero means the default limit
    value = int(value) or CONF.api_max_limit
    assert value >= 0, _('Limit cannot be negative')
    assert value <= CONF.api_max_limit, _('Limit over %s') % CONF.api_max_limit
    return value



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors

import ironic_inspector.conf
from ironic_inspector import version


MIN_VERSION_HEADER = 'X-OpenStack-Ironic-Inspector-API-Minimum-Version'
MAX_VERSION_HEADER = 'X-OpenStack-Ironic-Inspector-API-Maximum-Version'
VERSION_HEADER = 'X-OpenStack-Ironic-Inspector-API-Version'


def set_config_defaults():
    """Return a list of oslo.config options available in Inspector code."""
    log.set_defaults(default_log_levels=['sqlalchemy=WARNING',
                                         'iso8601=WARNING',
                                         'requests=WARNING',
                                         'urllib3.connectionpool=WARNING',
                                         'keystonemiddleware=WARNING',
                                         'keystoneauth=WARNING',
                                         'ironicclient=WARNING',
                                         'amqp=WARNING',
                                         'amqplib=WARNING',
                                         # This comes in two flavors
                                         'oslo.messaging=WARNING',
                                         'oslo_messaging=WARNING'])
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       MIN_VERSION_HEADER,
                       MAX_VERSION_HEADER,
                       VERSION_HEADER],
        allow_methods=['GET', 'POST', 'PUT', 'HEAD',
                       'PATCH', 'DELETE', 'OPTIONS']
    )


def parse_args(args, default_config_files=None):
    cfg.CONF(args,
             project='ironic-inspector',
             version=version.version_info.release_string(),
             default_config_files=default_config_files)


def list_opts():
    return [
        ('capabilities', ironic_inspector.conf.capabilities.list_opts()),
        ('coordination', ironic_inspector.conf.coordination.list_opts()),
        ('DEFAULT', ironic_inspector.conf.default.list_opts()),
        ('discovery', ironic_inspector.conf.discovery.list_opts()),
        ('dnsmasq_pxe_filter',
         ironic_inspector.conf.dnsmasq_pxe_filter.list_opts()),
        ('extra_hardware', ironic_inspector.conf.extra_hardware.list_opts()),
        ('ironic', ironic_inspector.conf.ironic.list_opts()),
        ('iptables', ironic_inspector.conf.iptables.list_opts()),
        ('port_physnet', ironic_inspector.conf.port_physnet.list_opts()),
        ('processing', ironic_inspector.conf.processing.list_opts()),
        ('pci_devices', ironic_inspector.conf.pci_devices.list_opts()),
        ('pxe_filter', ironic_inspector.conf.pxe_filter.list_opts()),
        ('service_catalog', ironic_inspector.conf.service_catalog.list_opts()),
        ('swift', ironic_inspector.conf.swift.list_opts()),
    ]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Introducing Node.state attribute

Revision ID: d2e48801c8ef
Revises: e169a4a81d88
Create Date: 2016-07-29 10:10:32.351661

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

from ironic_inspector import introspection_state as istate


# revision identifiers, used by Alembic.
revision = 'd2e48801c8ef'
down_revision = 'e169a4a81d88'
branch_labels = None
depends_on = None

Node = sql.table('nodes',
                 sql.column('error', sa.String),
                 sql.column('state', sa.Enum(*istate.States.all())))


def upgrade():
    state_enum = sa.Enum(*istate.States.all(), name='node_state')
    state_enum.create(op.get_bind())

    op.add_column('nodes', sa.Column('version_id', sa.String(36),
                                     server_default=''))
    op.add_column('nodes', sa.Column('state', state_enum,
                                     nullable=False,
                                     default=istate.States.finished,
                                     server_default=istate.States.finished))
    # correct the state: finished -> error if Node.error is not null
    stmt = Node.update().where(Node.c.error != sql.null()).values(
        {'state': op.inline_literal(istate.States.error)})
    op.execute(stmt)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add Rules

Revision ID: d588418040d
Revises: 578f84f38d
Create Date: 2015-09-21 14:31:03.048455

"""

from alembic import op
from oslo_db.sqlalchemy import types
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd588418040d'
down_revision = '578f84f38d'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'rules',
        sa.Column('uuid', sa.String(36), primary_key=True),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('description', sa.Text),
        sa.Column('disabled', sa.Boolean, default=False),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'rule_conditions',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('rule', sa.String(36), sa.ForeignKey('rules.uuid')),
        sa.Column('op', sa.String(255), nullable=False),
        sa.Column('multiple', sa.String(255), nullable=False),
        sa.Column('field', sa.Text),
        sa.Column('params', types.JsonEncodedDict),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'rule_actions',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('rule', sa.String(36), sa.ForeignKey('rules.uuid')),
        sa.Column('action', sa.String(255), nullable=False),
        sa.Column('params', types.JsonEncodedDict),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Enroll node not found hook hook."""

from oslo_config import cfg

from ironic_inspector.common import ironic as ir_utils
from ironic_inspector import node_cache
from ironic_inspector import utils


CONF = cfg.CONF

LOG = utils.getProcessingLogger(__name__)


def _extract_node_driver_info(introspection_data):
    node_driver_info = {}
    for ip_version in CONF.discovery.enabled_bmc_address_version:
        address = None
        if ip_version == '4':
            address = utils.get_ipmi_address_from_data(introspection_data)
        elif ip_version == '6':
            address = utils.get_ipmi_v6address_from_data(introspection_data)

        if address:
            node_driver_info['ipmi_address'] = address
            break
    else:
        LOG.warning('No BMC address provided, discovered node will be '
                    'created without ipmi address')
    return node_driver_info


def _check_existing_nodes(introspection_data, node_driver_info, ironic):
    macs = utils.get_valid_macs(introspection_data)
    if macs:
        ir_utils.lookup_node_by_macs(macs, introspection_data, ironic=ironic,
                                     fail=True)
    else:
        LOG.warning('No suitable interfaces found for discovered node. '
                    'Check that validate_interfaces hook is listed in '
                    '[processing]default_processing_hooks config option')

    # verify existing node with discovered ipmi address
    ipmi_address = node_driver_info.get('ipmi_address')
    if ipmi_address:
        ir_utils.lookup_node_by_bmc_addresses([ipmi_address],
                                              introspection_data,
                                              ironic=ironic, fail=True)


def enroll_node_not_found_hook(introspection_data, **kwargs):
    node_attr = CONF.discovery.enroll_node_fields.copy()
    ironic = ir_utils.get_client()

    node_driver_info = _extract_node_driver_info(introspection_data)
    node_attr['driver_info'] = node_driver_info

    # NOTE(rpittau) by default, if the provision_state is None, it will
    # be set to 'available' by openstacksdk, blocking the inspection of the
    # node in this phase, as it's not a valid state for inspection.
    node_attr['provision_state'] = 'enroll'

    node_driver = CONF.discovery.enroll_node_driver

    _check_existing_nodes(introspection_data, node_driver_info, ironic)

    LOG.debug('Creating discovered node with driver %(driver)s and '
              'attributes: %(attr)s',
              {'driver': node_driver, 'attr': node_attr},
              data=introspection_data)
    # NOTE(aarefiev): This flag allows to distinguish enrolled manually
    # and auto-discovered nodes in the introspection rules.
    introspection_data['auto_discovered'] = True
    return node_cache.create_node(node_driver, ironic=ironic, **node_attr)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Standard set of plugins."""

from ironic_lib import utils as il_utils
import netaddr
from oslo_config import cfg
from oslo_utils import netutils
from oslo_utils import units


from ironic_inspector.common.i18n import _
from ironic_inspector.common import ironic as ir_utils
from ironic_inspector.plugins import base
from ironic_inspector import utils

CONF = cfg.CONF

LOG = utils.getProcessingLogger('ironic_inspector.plugins.standard')


class RootDiskSelectionHook(base.ProcessingHook):
    """Smarter root disk selection using Ironic root device hints.

    This hook must always go before SchedulerHook, otherwise root_disk field
    might not be updated.
    """

    def _process_root_device_hints(self, introspection_data, node_info,
                                   inventory):
        """Detect root disk from root device hints and IPA inventory."""
        hints = node_info.node().properties.get('root_device')
        if not hints:
            LOG.debug('Root device hints are not provided',
                      node_info=node_info, data=introspection_data)
            return

        try:
            device = il_utils.match_root_device_hints(inventory['disks'],
                                                      hints)
        except (TypeError, ValueError) as e:
            raise utils.Error(
                _('No disks could be found using the root device hints '
                  '%(hints)s because they failed to validate. '
                  'Error: %(error)s') % {'hints': hints, 'error': e},
                node_info=node_info, data=introspection_data)

        if not device:
            raise utils.Error(_('No disks satisfied root device hints'),
                              node_info=node_info, data=introspection_data)

        LOG.debug('Disk %(disk)s of size %(size)s satisfies '
                  'root device hints',
                  {'disk': device.get('name'), 'size': device['size']},
                  node_info=node_info, data=introspection_data)
        introspection_data['root_disk'] = device

    def before_update(self, introspection_data, node_info, **kwargs):
        """Process root disk information."""
        inventory = utils.get_inventory(introspection_data,
                                        node_info=node_info)
        self._process_root_device_hints(introspection_data, node_info,
                                        inventory)

        root_disk = introspection_data.get('root_disk')
        if root_disk:
            local_gb = root_disk['size'] // units.Gi
            if not local_gb:
                LOG.warning('The requested root disk is too small (smaller '
                            'than 1 GiB) or its size cannot be detected: %s',
                            root_disk,
                            node_info=node_info, data=introspection_data)
            else:
                if CONF.processing.disk_partitioning_spacing:
                    local_gb -= 1
                LOG.info('Root disk %(disk)s, local_gb %(local_gb)s GiB',
                         {'disk': root_disk, 'local_gb': local_gb},
                         node_info=node_info, data=introspection_data)
        else:
            local_gb = 0
            LOG.info('No root device found, assuming a diskless node',
                     node_info=node_info, data=introspection_data)

        introspection_data['local_gb'] = local_gb
        if (CONF.processing.overwrite_existing or not
                node_info.node().properties.get('local_gb')):
            node_info.update_properties(local_gb=str(local_gb))


class SchedulerHook(base.ProcessingHook):
    """Nova scheduler required properties."""

    KEYS = ('cpus', 'cpu_arch', 'memory_mb')

    def before_update(self, introspection_data, node_info, **kwargs):
        """Update node with scheduler properties."""
        inventory = utils.get_inventory(introspection_data,
                                        node_info=node_info)
        try:
            introspection_data['cpus'] = int(inventory['cpu']['count'])
            introspection_data['cpu_arch'] = str(
                inventory['cpu']['architecture'])
        except (KeyError, ValueError, TypeError):
            LOG.warning('malformed or missing CPU information: %s',
                        inventory.get('cpu'))

        try:
            introspection_data['memory_mb'] = int(
                inventory['memory']['physical_mb'])
        except (KeyError, ValueError, TypeError):
            LOG.warning('malformed or missing memory information: %s; '
                        'introspection requires physical memory size '
                        'from dmidecode', inventory.get('memory'))

        LOG.info('Discovered data: CPUs: count %(cpus)s, architecture '
                 '%(cpu_arch)s, memory %(memory_mb)s MiB',
                 {key: introspection_data.get(key) for key in self.KEYS},
                 node_info=node_info, data=introspection_data)

        overwrite = CONF.processing.overwrite_existing
        properties = {key: str(introspection_data[key])
                      for key in self.KEYS if introspection_data.get(key) and
                      (overwrite or not node_info.node().properties.get(key))}
        if properties:
            node_info.update_properties(**properties)


class ValidateInterfacesHook(base.ProcessingHook):
    """Hook to validate network interfaces."""

    def __init__(self):
        # Some configuration checks
        if (CONF.processing.add_ports == 'disabled' and
                CONF.processing.keep_ports == 'added'):
            msg = _("Configuration error: add_ports set to disabled "
                    "and keep_ports set to added. Please change keep_ports "
                    "to all.")
            raise utils.Error(msg)

    def _get_interfaces(self, data=None):
        """Convert inventory to a dict with interfaces.

        :return: dict interface name -> dict with keys 'mac' and 'ip'
        """
        result = {}
        inventory = utils.get_inventory(data)

        pxe_mac = utils.get_pxe_mac(data)

        for iface in inventory['interfaces']:
            name = iface.get('name')
            mac = iface.get('mac_address')
            ipv4_address = iface.get('ipv4_address')
            ipv6_address = iface.get('ipv6_address')
            # NOTE(kaifeng) ipv6 address may in the form of fd00::1%enp2s0,
            # which is not supported by netaddr, remove the suffix if exists.
            if ipv6_address and '%' in ipv6_address:
                ipv6_address = ipv6_address.split('%')[0]
            ip = ipv4_address or ipv6_address
            client_id = iface.get('client_id')

            if not name:
                LOG.error('Malformed interface record: %s',
                          iface, data=data)
                continue

            if not mac:
                LOG.debug('Skipping interface %s without link information',
                          name, data=data)
                continue

            if not netutils.is_valid_mac(mac):
                LOG.warning('MAC %(mac)s for interface %(name)s is '
                            'not valid, skipping',
                            {'mac': mac, 'name': name},
                            data=data)
                continue

            mac = mac.lower()

            LOG.debug('Found interface %(name)s with MAC "%(mac)s", '
                      'IP address "%(ip)s" and client_id "%(client_id)s"',
                      {'name': name, 'mac': mac, 'ip': ip,
                       'client_id': client_id}, data=data)
            result[name] = {'ip': ip, 'mac': mac, 'client_id': client_id,
                            'pxe': (mac == pxe_mac)}

        return result

    def _validate_interfaces(self, interfaces, data=None):
        """Validate interfaces on correctness and suitability.

        :return: dict interface name -> dict with keys 'mac' and 'ip'
        """
        if not interfaces:
            raise utils.Error(_('No interfaces supplied by the ramdisk'),
                              data=data)

        pxe_mac = utils.get_pxe_mac(data)
        if not pxe_mac and CONF.processing.add_ports == 'pxe':
            LOG.warning('No boot interface provided in the introspection '
                        'data, will add all ports with IP addresses')

        result = {}

        for name, iface in interfaces.items():
            ip = iface.get('ip')
            pxe = iface.get('pxe', True)

            if name == 'lo' or (ip and netaddr.IPAddress(ip).is_loopback()):
                LOG.debug('Skipping local interface %s', name, data=data)
                continue

            if CONF.processing.add_ports == 'pxe' and pxe_mac and not pxe:
                LOG.debug('Skipping interface %s as it was not PXE booting',
                          name, data=data)
                continue
            elif CONF.processing.add_ports != 'all' and not ip:
                LOG.debug('Skipping interface %s as it did not have '
                          'an IP address assigned during the ramdisk run',
                          name, data=data)
                continue

            result[name] = iface

        if not result:
            raise utils.Error(_('No suitable interfaces found in %s') %
                              interfaces, data=data)
        return result

    def before_processing(self, introspection_data, **kwargs):
        """Validate information about network interfaces."""

        bmc_address = utils.get_ipmi_address_from_data(introspection_data)
        bmc_v6address = utils.get_ipmi_v6address_from_data(introspection_data)
        # Overwrite the old ipmi_address field to avoid inconsistency
        introspection_data['ipmi_address'] = bmc_address
        introspection_data['ipmi_v6address'] = bmc_v6address
        if not (bmc_address or bmc_v6address):
            LOG.debug('No BMC address provided in introspection data, '
                      'assuming virtual environment', data=introspection_data)

        all_interfaces = self._get_interfaces(introspection_data)

        interfaces = self._validate_interfaces(all_interfaces,
                                               introspection_data)

        LOG.info('Using network interface(s): %s',
                 ', '.join('%s %s' % (name, items)
                           for (name, items) in interfaces.items()),
                 data=introspection_data)

        introspection_data['all_interfaces'] = all_interfaces
        introspection_data['interfaces'] = interfaces
        valid_macs = [iface['mac'] for iface in interfaces.values()]
        introspection_data['macs'] = valid_macs

    def before_update(self, introspection_data, node_info, **kwargs):
        """Create new ports and drop ports that are not present in the data."""
        interfaces = introspection_data.get('interfaces')
        if CONF.processing.add_ports != 'disabled':
            node_info.create_ports(list(interfaces.values()))

        if CONF.processing.keep_ports == 'present':
            expected_macs = {
                iface['mac']
                for iface in introspection_data['all_interfaces'].values()
            }
        elif CONF.processing.keep_ports == 'added':
            expected_macs = set(introspection_data['macs'])

        node_provision_state = node_info.node().provision_state
        if node_provision_state in ir_utils.VALID_ACTIVE_STATES:
            LOG.debug('Node %(node)s is %(state)s; '
                      'not modifying / deleting its ports',
                      {'node': node_info.uuid,
                       'state': node_provision_state})
            return

        if CONF.processing.keep_ports != 'all':
            # list is required as we modify underlying dict
            for port in list(node_info.ports().values()):
                if port.address not in expected_macs:
                    LOG.info("Deleting port %(port)s as its MAC %(mac)s is "
                             "not in expected MAC list %(expected)s",
                             {'port': port.id,
                              'mac': port.address,
                              'expected': list(sorted(expected_macs))},
                             node_info=node_info, data=introspection_data)
                    node_info.delete_port(port)

        if (CONF.processing.overwrite_existing
                and CONF.processing.update_pxe_enabled):
            # Make sure is_pxe_enabled is up-to-date
            ports = node_info.ports()
            for iface in introspection_data['interfaces'].values():
                try:
                    port = ports[iface['mac']]
                except KeyError:
                    continue

                real_pxe = iface.get('pxe', True)
                if port.is_pxe_enabled != real_pxe:
                    LOG.info('Fixing is_pxe_enabled=%(val)s on port %(port)s '
                             'to match introspected data',
                             {'port': port.address, 'val': real_pxe},
                             node_info=node_info, data=introspection_data)
                    node_info.patch_port(port, [{'op': 'replace',
                                                 'path': '/pxe_enabled',
                                                 'value': real_pxe}])


class RamdiskErrorHook(base.ProcessingHook):
    """Hook to process error send from the ramdisk."""

    def before_processing(self, introspection_data, **kwargs):
        error = introspection_data.get('error')
        if error:
            raise utils.Error(_('Ramdisk reported error: %s') % error,
                              data=introspection_data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
try:
    import errno
except ImportError:
    import os.errno as errno
import datetime
import os
from unittest import mock

import fixtures
from openstack import exceptions as os_exc
from oslo_config import cfg

from ironic_inspector.common import ironic as ir_utils
from ironic_inspector import node_cache
from ironic_inspector.pxe_filter import base as pxe_filter
from ironic_inspector.pxe_filter import dnsmasq
from ironic_inspector.test import base as test_base
from ironic_inspector import utils

CONF = cfg.CONF


class DnsmasqTestBase(test_base.BaseTest):
    def setUp(self):
        super(DnsmasqTestBase, self).setUp()
        self.driver = dnsmasq.DnsmasqFilter()


class TestConfiguration(DnsmasqTestBase):

    def test_deny_unknown_macs_and_node_not_found_hook_bad(self):
        CONF.set_override('deny_unknown_macs', True, 'pxe_filter')
        CONF.set_override('node_not_found_hook', True, 'processing')
        self.assertRaisesRegex(utils.Error, 'Configuration error:',
                               self.driver.__init__)


class TestShouldEnableUnknownHosts(DnsmasqTestBase):
    def setUp(self):
        super(TestShouldEnableUnknownHosts, self).setUp()
        self.mock_introspection_active = self.useFixture(
            fixtures.MockPatchObject(node_cache, 'introspection_active')).mock

    def test_introspection_active(self):
        self.mock_introspection_active.return_value = True
        self.assertTrue(dnsmasq._should_enable_unknown_hosts())

    def test_introspection_not_active(self):
        self.mock_introspection_active.return_value = False
        self.assertFalse(dnsmasq._should_enable_unknown_hosts())


class TestDnsmasqDriverAPI(DnsmasqTestBase):
    def setUp(self):
        super(TestDnsmasqDriverAPI, self).setUp()
        self.mock__execute = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_execute')).mock
        self.driver._sync = mock.Mock()
        self.driver._tear_down = mock.Mock()
        self.mock__purge_dhcp_hostsdir = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_purge_dhcp_hostsdir')).mock
        self.mock_ironic = mock.Mock()
        get_client_mock = self.useFixture(
            fixtures.MockPatchObject(ir_utils, 'get_client')).mock
        get_client_mock.return_value = self.mock_ironic
        self.start_command = '/far/boo buzz -V --ack 42'
        CONF.set_override('dnsmasq_start_command', self.start_command,
                          'dnsmasq_pxe_filter')
        self.stop_command = '/what/ever'
        CONF.set_override('dnsmasq_stop_command', self.stop_command,
                          'dnsmasq_pxe_filter')

    def test_init_filter(self):
        self.driver.init_filter()

        self.mock__purge_dhcp_hostsdir.assert_called_once_with()
        self.driver._sync.assert_called_once_with(self.mock_ironic)
        self.mock__execute.assert_called_once_with(self.start_command)

    def test_sync(self):
        self.driver.init_filter()
        # NOTE(milan) init_filter performs an initial sync
        self.driver._sync.reset_mock()
        self.driver.sync(self.mock_ironic)

        self.driver._sync.assert_called_once_with(self.mock_ironic)

    def test_tear_down_filter(self):
        mock_reset = self.useFixture(
            fixtures.MockPatchObject(self.driver, 'reset')).mock
        self.driver.init_filter()
        self.driver.tear_down_filter()

        mock_reset.assert_called_once_with()

    def test_reset(self):
        self.driver.init_filter()
        # NOTE(milan) init_filter calls _base_cmd
        self.mock__execute.reset_mock()
        self.driver.reset()

        self.mock__execute.assert_called_once_with(
            self.stop_command, ignore_errors=True)


class TestExclusiveWriteOrPass(test_base.BaseTest):
    def setUp(self):
        super(TestExclusiveWriteOrPass, self).setUp()
        self.mock_open = self.useFixture(fixtures.MockPatchObject(
            builtins, 'open', new=mock.mock_open())).mock
        self.mock_fd = self.mock_open.return_value
        self.mock_fcntl = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.fcntl, 'flock', autospec=True)).mock
        self.path = '/foo/bar/baz'
        self.buf = 'spam'
        self.fcntl_lock_call = mock.call(
            self.mock_fd, dnsmasq.fcntl.LOCK_EX | dnsmasq.fcntl.LOCK_NB)
        self.fcntl_unlock_call = mock.call(self.mock_fd, dnsmasq.fcntl.LOCK_UN)
        self.mock_log = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.LOG, 'debug')).mock
        self.mock_sleep = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.time, 'sleep')).mock

    def test_write(self):
        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)
        self.assertTrue(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_called_once_with(self.buf)
        self.mock_log.assert_not_called()

    def test_write_would_block(self):
        err = IOError('Oops!')
        err.errno = errno.EWOULDBLOCK
        # lock/unlock paired calls
        self.mock_fcntl.side_effect = [
            # first try
            err, None,
            # second try
            None, None]
        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)

        self.assertTrue(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call],
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_called_once_with(self.buf)
        self.mock_log.assert_called_once_with(
            '%s locked; will try again (later)', self.path)
        self.mock_sleep.assert_called_once_with(
            dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS_DELAY)

    def test_write_would_block_too_many_times(self):
        self.useFixture(fixtures.MonkeyPatch(
            'ironic_inspector.pxe_filter.dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS',
            1))
        err = IOError('Oops!')
        err.errno = errno.EWOULDBLOCK
        self.mock_fcntl.side_effect = [err, None]

        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)
        self.assertFalse(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_not_called()
        retry_log_call = mock.call('%s locked; will try again (later)',
                                   self.path)
        failed_log_call = mock.call(
            'Failed to write the exclusively-locked path: %(path)s for '
            '%(attempts)s times', {
                'attempts': dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS,
                'path': self.path
            })
        self.mock_log.assert_has_calls([retry_log_call, failed_log_call])
        self.mock_sleep.assert_called_once_with(
            dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS_DELAY)

    def test_write_custom_ioerror(self):

        err = IOError('Oops!')
        err.errno = errno.EBADF
        self.mock_fcntl.side_effect = [err, None]

        self.assertRaisesRegex(
            IOError, 'Oops!', dnsmasq._exclusive_write_or_pass, self.path,
            self.buf)

        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_not_called()
        self.mock_log.assert_not_called()


class TestMACHandlers(test_base.BaseTest):
    def setUp(self):
        super(TestMACHandlers, self).setUp()
        self.mac = 'ff:ff:ff:ff:ff:ff'
        self.dhcp_hostsdir = '/far'
        CONF.set_override('dhcp_hostsdir', self.dhcp_hostsdir,
                          'dnsmasq_pxe_filter')
        self.mock_join = self.useFixture(
            fixtures.MockPatchObject(os.path, 'join')).mock
        self.mock_join.return_value = "%s/%s" % (self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_exclusive_write_or_pass')).mock
        self.mock_stat = self.useFixture(
            fixtures.MockPatchObject(os, 'stat')).mock
        self.mock_listdir = self.useFixture(
            fixtures.MockPatchObject(os, 'listdir')).mock
        self.mock_remove = self.useFixture(
            fixtures.MockPatchObject(os, 'remove')).mock
        self.mock_log = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, 'LOG')).mock
        self.mock_introspection_active = self.useFixture(
            fixtures.MockPatchObject(node_cache, 'introspection_active')).mock

    def test__allowlist_unknown_hosts(self):
        self.mock_join.return_value = "%s/%s" % (self.dhcp_hostsdir,
                                                 dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock_introspection_active.return_value = True
        dnsmasq._configure_unknown_hosts()

        self.mock_join.assert_called_once_with(self.dhcp_hostsdir,
                                               dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value,
            '%s' % dnsmasq._ALLOW_UNKNOWN_HOSTS)
        self.mock_log.debug.assert_called_once_with(
            'A %s record for all unknown hosts using wildcard mac '
            'created', 'allow')

    def test__denylist_unknown_hosts(self):
        self.mock_join.return_value = "%s/%s" % (self.dhcp_hostsdir,
                                                 dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock_introspection_active.return_value = False
        dnsmasq._configure_unknown_hosts()

        self.mock_join.assert_called_once_with(self.dhcp_hostsdir,
                                               dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value,
            '%s' % dnsmasq._DENY_UNKNOWN_HOSTS)
        self.mock_log.debug.assert_called_once_with(
            'A %s record for all unknown hosts using wildcard mac '
            'created', 'deny')

    def test__macs_unknown_hosts_deny_unknown(self):
        self.mock_join.return_value = "%s/%s" % (self.dhcp_hostsdir,
                                                 dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock_introspection_active.return_value = True
        CONF.set_override('deny_unknown_macs', True, 'pxe_filter')

        dnsmasq._configure_unknown_hosts()

        self.mock_join.assert_called_once_with(self.dhcp_hostsdir,
                                               dnsmasq._UNKNOWN_HOSTS_FILE)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value,
            '%s' % dnsmasq._DENY_UNKNOWN_HOSTS)
        self.mock_log.debug.assert_called_once_with(
            'A %s record for all unknown hosts using wildcard mac '
            'created', 'deny')

    def test__configure_removedlist_allowlist(self):
        self.mock_introspection_active.return_value = True
        self.mock_stat.return_value.st_size = dnsmasq._MAC_DENY_LEN

        dnsmasq._configure_removedlist({self.mac})

        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value, '%s\n' % self.mac)

    def test__configure_removedlist_denylist(self):
        self.mock_introspection_active.return_value = False
        self.mock_stat.return_value.st_size = dnsmasq._MAC_ALLOW_LEN

        dnsmasq._configure_removedlist({self.mac})

        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value, '%s,ignore\n' % self.mac)

    def test__configure_removedlist_denylist_deny_unknown(self):
        self.mock_introspection_active.return_value = True
        self.mock_stat.return_value.st_size = dnsmasq._MAC_ALLOW_LEN
        CONF.set_override('deny_unknown_macs', True, 'pxe_filter')

        dnsmasq._configure_removedlist({self.mac})

        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value, '%s,ignore\n' % self.mac)

    def test__allowlist_mac(self):
        dnsmasq._add_mac_to_allowlist(self.mac)

        self.mock_join.assert_called_once_with(self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value, '%s\n' % self.mac)

    def test__denylist_mac(self):
        dnsmasq._add_mac_to_denylist(self.mac)

        self.mock_join.assert_called_once_with(self.dhcp_hostsdir, self.mac)
        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.mock_join.return_value, '%s,ignore\n' % self.mac)

    def test__get_denylist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s,ignore\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual({self.mac}, denylist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock_stat.assert_called_with(self.mock_join.return_value)

    def test__get_allowlist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual({self.mac}, allowlist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock_stat.assert_called_with(self.mock_join.return_value)

    def test__get_no_denylist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual(set(), denylist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock_stat.assert_called_with(self.mock_join.return_value)

    def test__get_no_allowlist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s,ignore\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual(set(), allowlist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_join.assert_called_with(self.dhcp_hostsdir, self.mac)
        self.mock_stat.assert_called_with(self.mock_join.return_value)

    def test__purge_dhcp_hostsdir(self):
        self.mock_listdir.return_value = [self.mac]
        dnsmasq._purge_dhcp_hostsdir()

        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_join.assert_called_once_with(self.dhcp_hostsdir, self.mac)
        self.mock_remove.assert_called_once_with('%s/%s' % (self.dhcp_hostsdir,
                                                            self.mac))

    def test_disabled__purge_dhcp_hostsdir(self):
        CONF.set_override('purge_dhcp_hostsdir', False, 'dnsmasq_pxe_filter')
        # NOTE(dtantsur): set_override uses os.path internally
        self.mock_join.reset_mock()

        dnsmasq._purge_dhcp_hostsdir()
        self.mock_listdir.assert_not_called()
        self.mock_join.assert_not_called()
        self.mock_remove.assert_not_called()


class TestSync(DnsmasqTestBase):
    def setUp(self):
        super(TestSync, self).setUp()
        self.mock__get_deny_allow_lists = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_get_deny_allow_lists')).mock
        self.mock__add_mac_to_allowlist = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_add_mac_to_allowlist')).mock
        self.mock__add_mac_to_denylist = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_add_mac_to_denylist')).mock
        self.mock__configure_unknown_hosts = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_configure_unknown_hosts')).mock
        self.mock__configure_removedlist = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_configure_removedlist')).mock

        self.mock_ironic = mock.Mock()
        self.mock_utcnow = self.useFixture(
            fixtures.MockPatchObject(dnsmasq.timeutils, 'utcnow')).mock
        self.timestamp_start = datetime.datetime.utcnow()
        self.timestamp_end = (self.timestamp_start +
                              datetime.timedelta(seconds=42))
        self.mock_utcnow.side_effect = [self.timestamp_start,
                                        self.timestamp_end]
        self.mock_log = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, 'LOG')).mock
        get_client_mock = self.useFixture(
            fixtures.MockPatchObject(ir_utils, 'get_client')).mock
        get_client_mock.return_value = self.mock_ironic
        self.mock_get_active_macs = self.useFixture(
            fixtures.MockPatchObject(pxe_filter, 'get_active_macs')).mock
        self.mock_get_active_macs.return_value = set()
        self.mock_active_macs = self.useFixture(
            fixtures.MockPatchObject(node_cache, 'active_macs')).mock
        self.ironic_macs = {'new_mac', 'active_mac'}
        self.active_macs = {'active_mac'}
        self.denylist = {'gone_mac', 'active_mac'}
        self.allowlist = {}
        self.mock__get_deny_allow_lists.return_value = (self.denylist,
                                                        self.allowlist)
        self.mock_ironic.ports.return_value = [
            mock.Mock(address=address) for address in self.ironic_macs]
        self.mock_active_macs.return_value = self.active_macs
        self.mock_should_enable_unknown_hosts = self.useFixture(
            fixtures.MockPatchObject(dnsmasq,
                                     '_should_enable_unknown_hosts')).mock
        self.mock_should_enable_unknown_hosts.return_value = True

    def test__sync_enable_unknown_hosts(self):
        self.mock_should_enable_unknown_hosts.return_value = True

        self.driver._sync(self.mock_ironic)
        self.mock__configure_unknown_hosts.assert_called_once_with()

    def test__sync_not_enable_unknown_hosts(self):
        self.mock_should_enable_unknown_hosts.return_value = False

        self.driver._sync(self.mock_ironic)
        self.mock__configure_unknown_hosts.assert_called_once_with()

    def test__sync(self):
        self.mock_get_active_macs.return_value = {'active_mac'}
        self.driver._sync(self.mock_ironic)

        self.mock_get_active_macs.assert_called_once_with(self.mock_ironic)
        self.mock__add_mac_to_allowlist.assert_called_once_with('active_mac')
        self.mock__add_mac_to_denylist.assert_called_once_with('new_mac')

        self.mock_ironic.ports.assert_called_once_with(
            fields=['address', 'extra'], limit=None)
        self.mock__get_deny_allow_lists.assert_called_once_with()
        self.mock__configure_unknown_hosts.assert_called_once_with()
        self.mock__configure_removedlist.assert_called_once_with({'gone_mac'})
        self.mock_log.debug.assert_has_calls([
            mock.call('Syncing the driver'),
            mock.call('The dnsmasq PXE filter was synchronized (took %s)',
                      self.timestamp_end - self.timestamp_start)
        ])

    @mock.patch('time.sleep', lambda _x: None)
    def test__sync_with_port_list_retries(self):
        self.mock_ironic.ports.side_effect = [
            os_exc.SDKException('boom'),
            [mock.Mock(address=address) for address in self.ironic_macs]
        ]
        self.mock_get_active_macs.return_value = {'active_mac'}
        self.driver._sync(self.mock_ironic)

        self.mock__add_mac_to_allowlist.assert_called_once_with('active_mac')
        self.mock__add_mac_to_denylist.assert_called_once_with('new_mac')

        self.mock_ironic.ports.assert_called_with(fields=['address', 'extra'],
                                                  limit=None)
        self.mock_get_active_macs.assert_called_once_with(self.mock_ironic)
        self.mock__get_deny_allow_lists.assert_called_once_with()
        self.mock__configure_removedlist.assert_called_once_with({'gone_mac'})
        self.mock_log.debug.assert_has_calls([
            mock.call('Syncing the driver'),
            mock.call('The dnsmasq PXE filter was synchronized (took %s)',
                      self.timestamp_end - self.timestamp_start)
        ])


class Test_Execute(test_base.BaseTest):
    def setUp(self):
        super(Test_Execute, self).setUp()
        self.mock_execute = self.useFixture(
            fixtures.MockPatchObject(dnsmasq.processutils, 'execute')
        ).mock
        CONF.set_override('rootwrap_config', '/path/to/rootwrap.conf')
        self.rootwrap_cmd = dnsmasq._ROOTWRAP_COMMAND.format(
            rootwrap_config=CONF.rootwrap_config)
        self.useFixture(fixtures.MonkeyPatch(
            'ironic_inspector.pxe_filter.dnsmasq._ROOTWRAP_COMMAND',
            self.rootwrap_cmd))
        self.command = 'foobar baz'

    def test__execute(self):
        dnsmasq._execute(self.command)
        self.mock_execute.assert_called_once_with(
            self.command, run_as_root=True, shell=True,
            check_exit_code=True, root_helper=self.rootwrap_cmd)

    def test__execute_ignoring_errors(self):
        dnsmasq._execute(self.command, ignore_errors=True)
        self.mock_execute.assert_called_once_with(
            self.command, run_as_root=True, shell=True,
            check_exit_code=False, root_helper=self.rootwrap_cmd)

    def test__execute_empty(self):
        dnsmasq._execute()

        self.mock_execute.assert_not_called()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging as pylog

import futurist
from ironic_lib import auth_basic
from ironic_lib import exception
from keystonemiddleware import auth_token
from openstack.baremetal.v1 import node
from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors as cors_middleware
import pytz
import webob

from ironic_inspector.common.i18n import _
from ironic_inspector import policy

CONF = cfg.CONF

_EXECUTOR = None


def get_ipmi_address_from_data(introspection_data):
    try:
        result = introspection_data['inventory']['bmc_address']
    except KeyError:
        result = introspection_data.get('ipmi_address')

    if result in ('', '0.0.0.0'):
        # ipmitool can return these values, if it does not know the address
        return None
    else:
        return result


def get_ipmi_v6address_from_data(introspection_data):
    try:
        result = introspection_data['inventory']['bmc_v6address']
    except KeyError:
        result = introspection_data.get('ipmi_v6address')

    if result in ('', '::/0'):
        # ipmitool can return these values, if it does not know the address
        return None
    else:
        return result


def get_pxe_mac(introspection_data):
    pxe_mac = introspection_data.get('boot_interface')
    if pxe_mac and '-' in pxe_mac:
        # pxelinux format: 01-aa-bb-cc-dd-ee-ff
        pxe_mac = pxe_mac.split('-', 1)[1]
        pxe_mac = pxe_mac.replace('-', ':').lower()
    return pxe_mac


def processing_logger_prefix(data=None, node_info=None):
    """Calculate prefix for logging.

    Tries to use:
    * node UUID, node._state
    * node PXE MAC,
    * node BMC address

    :param data: introspection data
    :param node_info: NodeInfo or ironic node object
    :return: logging prefix as a string
    """
    # TODO(dtantsur): try to get MAC and BMC address for node_info as well
    parts = []
    data = data or {}

    if node_info is not None:
        if isinstance(node_info, node.Node):
            parts.append(str(node_info.id))
        else:
            parts.append(str(node_info))

    pxe_mac = get_pxe_mac(data)
    if pxe_mac:
        parts.append('MAC %s' % pxe_mac)

    bmc_address = get_ipmi_address_from_data(data) if data else None
    if bmc_address:
        parts.append('BMC %s' % bmc_address)

    if parts:
        return _('[node: %s]') % ' '.join(parts)
    else:
        return _('[unidentified node]')


class ProcessingLoggerAdapter(log.KeywordArgumentAdapter):
    def process(self, msg, kwargs):
        if 'data' not in kwargs and 'node_info' not in kwargs:
            return super(ProcessingLoggerAdapter, self).process(msg, kwargs)

        data = kwargs.get('data', {})
        node_info = kwargs.get('node_info')
        prefix = processing_logger_prefix(data, node_info)

        msg, kwargs = super(ProcessingLoggerAdapter, self).process(msg, kwargs)
        return ('%s %s' % (prefix, msg)), kwargs


def getProcessingLogger(name):
    # We can't use getLogger from oslo_log, as it's an adapter itself
    logger = pylog.getLogger(name)
    return ProcessingLoggerAdapter(logger, {})


LOG = getProcessingLogger(__name__)


class Error(Exception):
    """Inspector exception."""

    def __init__(self, msg, code=400, log_level='error', **kwargs):
        super(Error, self).__init__(msg)
        getattr(LOG, log_level)(msg, **kwargs)
        self.http_code = code


class NotFoundInCacheError(Error):
    """Exception when node was not found in cache during processing."""

    def __init__(self, msg, code=404, **kwargs):
        super(NotFoundInCacheError, self).__init__(msg, code,
                                                   log_level='info', **kwargs)


class NodeStateRaceCondition(Error):
    """State mismatch between the DB and a node_info."""
    def __init__(self, *args, **kwargs):
        message = _('Node state mismatch detected between the DB and the '
                    'cached node_info object')
        kwargs.setdefault('code', 500)
        super(NodeStateRaceCondition, self).__init__(message, *args, **kwargs)


class NodeStateInvalidEvent(Error):
    """Invalid event attempted."""


class IntrospectionDataStoreDisabled(Error):
    """Introspection data store is disabled."""


class IntrospectionDataNotFound(NotFoundInCacheError):
    """Introspection data not found."""


class NoAvailableConductor(Error):
    """No available conductor in the service group."""

    def __init__(self, msg, **kwargs):
        super(NoAvailableConductor, self).__init__(msg, code=503, **kwargs)


class DeferredBasicAuthMiddleware(object):
    """Middleware which sets X-Identity-Status header based on authentication

    """
    def __init__(self, app, auth_file):
        self.app = app
        self.auth_file = auth_file
        auth_basic.validate_auth_file(auth_file)

    @webob.dec.wsgify()
    def __call__(self, req):

        headers = req.headers
        try:
            if 'Authorization' not in headers:
                auth_basic.unauthorized()

            token = auth_basic.parse_header({
                'HTTP_AUTHORIZATION': headers.get('Authorization')
            })
            username, password = auth_basic.parse_token(token)
            headers.update(
                auth_basic.authenticate(self.auth_file, username, password))
            headers['X-Identity-Status'] = 'Confirmed'

        except exception.Unauthorized:
            headers['X-Identity-Status'] = 'Invalid'
        except exception.IronicException as e:
            status = '%s %s' % (int(e.code), str(e))
            resp = webob.Response(status=status)
            resp.headers.update(e.headers)
            return resp

        return req.get_response(self.app)


def executor():
    """Return the current futures executor."""
    global _EXECUTOR
    if _EXECUTOR is None:
        _EXECUTOR = futurist.GreenThreadPoolExecutor(
            max_workers=CONF.max_concurrency)
    return _EXECUTOR


def add_auth_middleware(app):
    """Add authentication middleware to Flask application.

    :param app: application.
    """
    auth_conf = dict(CONF.keystone_authtoken)
    auth_conf['delay_auth_decision'] = True
    app.wsgi_app = auth_token.AuthProtocol(app.wsgi_app, auth_conf)


def add_basic_auth_middleware(app):
    """Add HTTP Basic authentication middleware to Flask application.

    :param app: application.
    """
    app.wsgi_app = DeferredBasicAuthMiddleware(
        app.wsgi_app, CONF.http_basic_auth_user_file)


def add_cors_middleware(app):
    """Create a CORS wrapper

    Attach ironic-inspector-specific defaults that must be included
    in all CORS responses.

    :param app: application
    """
    app.wsgi_app = cors_middleware.CORS(app.wsgi_app, CONF)


def check_auth(request, rule=None, target=None):
    """Check authentication on request.

    :param request: Flask request
    :param rule: policy rule to check the request against
    :param target: dict-like structure to check rule against
    :raises: utils.Error if access is denied
    """
    if CONF.auth_strategy not in ('keystone', 'http_basic'):
        return
    if not request.context.is_public_api:
        if request.headers.get('X-Identity-Status', '').lower() == 'invalid':
            raise Error(_('Authentication required'), code=401)
    if CONF.auth_strategy != 'keystone':
        return
    target = {} if target is None else target
    if not policy.authorize(rule, target, request.context.to_policy_values()):
        raise Error(_("Access denied by policy"), code=403)


def get_valid_macs(data):
    """Get a list of valid MAC's from the introspection data."""
    return [m['mac']
            for m in data.get('all_interfaces', {}).values()
            if m.get('mac')]


def get_inventory(data, node_info=None):
    """Get and validate the hardware inventory from introspection data."""
    inventory = data.get('inventory')
    # TODO(dtantsur): validate inventory using JSON schema
    if not inventory:
        raise Error(_('Hardware inventory is empty or missing'),
                    data=data, node_info=node_info)

    if not inventory.get('interfaces'):
        raise Error(_('No network interfaces provided in the inventory'),
                    data=data, node_info=node_info)

    if not inventory.get('disks'):
        LOG.info('No disks were detected in the inventory, assuming this '
                 'is a disk-less node', data=data, node_info=node_info)
        # Make sure the code iterating over it does not fail with a TypeError
        inventory['disks'] = []

    return inventory


def iso_timestamp(timestamp=None, tz=pytz.timezone('utc')):
    """Return an ISO8601-formatted timestamp (tz: UTC) or None.

    :param timestamp: such as time.time() or None
    :param tz: timezone
    :returns: an ISO8601-formatted timestamp, or None
    """
    if timestamp is None:
        return None
    date = datetime.datetime.fromtimestamp(timestamp, tz=tz)
    return date.isoformat()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import fixtures
from oslo_config import cfg
import tooz

from ironic_inspector.common import coordination
from ironic_inspector.test import base
from ironic_inspector import utils

CONF = cfg.CONF


@mock.patch.object(coordination, 'Coordinator', autospec=True)
class TestGetCoordinator(base.BaseTest):
    def setUp(self):
        super(TestGetCoordinator, self).setUp()
        coordination._COORDINATOR = None

    def test_get(self, mock_coordinator):

        coordination.get_coordinator()
        mock_coordinator.assert_called_once_with(prefix=None)

    def test_get_with_prefix(self, mock_coordinator):
        coordination.get_coordinator(prefix='conductor')
        mock_coordinator.assert_called_once_with(prefix='conductor')


class TestCoordinator(base.BaseTest):
    def setUp(self):
        super(TestCoordinator, self).setUp()
        self.coordinator = coordination.Coordinator(prefix='test')
        self.mock_driver = self.useFixture(
            fixtures.MockPatchObject(tooz.coordination, 'CoordinationDriver',
                                     autospec=True)).mock
        self.mock_get_coordinator = self.useFixture(
            fixtures.MockPatchObject(tooz.coordination, 'get_coordinator',
                                     autospec=True)).mock
        self.mock_get_coordinator.return_value = self.mock_driver
        self.group_name = coordination.COORDINATION_GROUP_NAME.encode('ascii')

    def test_start(self):
        CONF.set_override('backend_url', 'memcached://1.2.3.4:11211',
                          'coordination')
        CONF.set_override('host', '1.2.3.5')
        self.coordinator.start()
        self.mock_get_coordinator.assert_called_once_with(
            'memcached://1.2.3.4:11211', b'ironic_inspector.test.1.2.3.5')
        self.assertTrue(self.coordinator.started)
        self.mock_driver.start.assert_called_once_with(start_heart=True)

    def test_stop(self):
        self.coordinator.started = True
        self.coordinator.coordinator = mock.MagicMock()
        self.coordinator.stop()
        self.assertFalse(self.coordinator.started)

    def test__create_group(self):
        self.coordinator.start()
        self.coordinator._create_group()
        self.mock_driver.create_group.assert_called_once_with(self.group_name)

    def test_join_group(self):
        self.coordinator.start()
        self.coordinator.join_group()
        self.mock_driver.join_group.assert_called_once_with(self.group_name)

    def test_join_group_not_exist(self):
        self.coordinator.start()
        self.mock_driver.join_group.side_effect = [
            tooz.coordination.GroupNotCreated('a group'), mock.Mock()]
        self.coordinator.join_group()
        self.mock_driver.create_group.assert_called_once_with(self.group_name)
        self.mock_driver.join_group.assert_has_calls([
            mock.call(self.group_name), mock.call(self.group_name)])

    def test_leave_group(self):
        self.coordinator.start()
        self.coordinator.leave_group()
        self.mock_driver.leave_group.assert_called_once_with(self.group_name)

    def test_get_members(self):
        self.coordinator.start()
        mock_resp = mock.Mock()
        mock_resp.get.return_value = {'host1', 'host2'}
        self.mock_driver.get_members.return_value = mock_resp
        members = self.coordinator.get_members()
        self.assertEqual(members, {'host1', 'host2'})
        self.mock_driver.get_members.assert_called_once_with(self.group_name)

    def test_get_members_no_such_group(self):
        self.coordinator.start()
        self.mock_driver.get_members.side_effect = (
            tooz.coordination.GroupNotCreated('a group'))
        self.assertEqual(self.coordinator.get_members(), set())

    def test_get_lock(self):
        self.coordinator.start()
        self.coordinator.get_lock('fake-node')
        self.mock_driver.get_lock.assert_called_once_with(
            b'ironic_inspector.fake-node')

    def test_invalid_state(self):
        self.assertRaisesRegex(utils.Error, 'Coordinator should be started',
                               self.coordinator.join_group)
        self.assertRaisesRegex(utils.Error, 'Coordinator should be started',
                               self.coordinator.leave_group)
        self.assertRaisesRegex(utils.Error, 'Coordinator should be started',
                               self.coordinator.get_members)
        self.assertRaisesRegex(utils.Error, 'Coordinator should be started',
                               self.coordinator.get_lock, 'fake id')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic_inspector.common.i18n import _


_OPTS = [
    cfg.StrOpt('dnsmasq_interface',
               default='br-ctlplane',
               help=_('Interface on which dnsmasq listens, the default is for '
                      'VM\'s.')),
    cfg.StrOpt('firewall_chain',
               default='ironic-inspector',
               help=_('iptables chain name to use.')),
    cfg.ListOpt('ethoib_interfaces',
                default=[],
                help=_('List of Ethernet Over InfiniBand interfaces '
                       'on the Inspector host which are used for physical '
                       'access to the DHCP network. Multiple interfaces would '
                       'be attached to a bond or bridge specified in '
                       'dnsmasq_interface. The MACs of the InfiniBand nodes '
                       'which are not in desired state are going to be '
                       'blacklisted based on the list of neighbor MACs '
                       'on these interfaces.')),
    cfg.StrOpt('ip_version',
               default='4',
               choices=[('4', _('IPv4')),
                        ('6', _('IPv6'))],
               help=_('The IP version that will be used for iptables filter. '
                      'Defaults to 4.')),
]


def register_opts(conf):
    conf.register_opts(_OPTS, 'iptables')


def list_opts():
    return _OPTS



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from unittest import mock

import flask
from oslo_config import cfg
from oslo_utils import uuidutils

from ironic_inspector import api_tools
import ironic_inspector.test.base as test_base
from ironic_inspector import utils

CONF = cfg.CONF
app = flask.Flask(__name__)
app.testing = True


def mock_test_field(return_value=None, side_effect=None):
    """Mock flask.request.args.get"""

    def outer(func):
        @functools.wraps(func)
        def inner(self, *args, **kwargs):
            with app.test_request_context('/'):
                get_mock = flask.request.args.get = mock.Mock()
                get_mock.return_value = return_value
                get_mock.side_effect = side_effect
                ret = func(self, get_mock, *args, **kwargs)
            return ret

        return inner

    return outer


class RaisesCoercionExceptionTestCase(test_base.BaseTest):
    def test_ok(self):
        @api_tools.raises_coercion_exceptions
        def fn():
            return True

        self.assertIs(True, fn())

    def test_assertion_error(self):
        @api_tools.raises_coercion_exceptions
        def fn():
            assert False, 'Oops!'

        self.assertRaisesRegex(utils.Error, 'Bad request: Oops!', fn)

    def test_value_error(self):
        @api_tools.raises_coercion_exceptions
        def fn():
            raise ValueError('Oops!')

        self.assertRaisesRegex(utils.Error, 'Bad request: Oops!', fn)


class RequestFieldTestCase(test_base.BaseTest):
    @mock_test_field(return_value='42')
    def test_request_field_ok(self, get_mock):
        @api_tools.request_field('foo')
        def fn(value):
            self.assertEqual(get_mock.return_value, value)

        fn()
        get_mock.assert_called_once_with('foo', default=None)

    @mock_test_field(return_value='42')
    def test_request_field_with_default(self, get_mock):
        @api_tools.request_field('foo')
        def fn(value):
            self.assertEqual(get_mock.return_value, value)

        fn(default='bar')
        get_mock.assert_called_once_with('foo', default='bar')

    @mock_test_field(return_value=42)
    def test_request_field_with_default_returns_default(self, get_mock):
        @api_tools.request_field('foo')
        def fn(value):
            self.assertEqual(get_mock.return_value, value)

        fn(default=42)
        get_mock.assert_called_once_with('foo', default=42)


class MarkerFieldTestCase(test_base.BaseTest):
    @mock_test_field(return_value=uuidutils.generate_uuid())
    def test_marker_ok(self, get_mock):
        value = api_tools.marker_field()
        self.assertEqual(get_mock.return_value, value)

    @mock.patch.object(uuidutils, 'is_uuid_like', autospec=True)
    @mock_test_field(return_value='foo')
    def test_marker_check_fails(self, get_mock, like_mock):
        like_mock.return_value = False
        self.assertRaisesRegex(utils.Error,
                               '.*(Marker not UUID-like)',
                               api_tools.marker_field)
        like_mock.assert_called_once_with(get_mock.return_value)


class LimitFieldTestCase(test_base.BaseTest):
    @mock_test_field(return_value=42)
    def test_limit_ok(self, get_mock):
        value = api_tools.limit_field()
        self.assertEqual(get_mock.return_value, value)

    @mock_test_field(return_value=str(CONF.api_max_limit + 1))
    def test_limit_over(self, get_mock):
        self.assertRaisesRegex(utils.Error,
                               '.*(Limit over %s)' % CONF.api_max_limit,
                               api_tools.limit_field)

    @mock_test_field(return_value='0')
    def test_limit_zero(self, get_mock):
        value = api_tools.limit_field()
        self.assertEqual(CONF.api_max_limit, value)

    @mock_test_field(return_value='-1')
    def test_limit_negative(self, get_mock):
        self.assertRaisesRegex(utils.Error,
                               '.*(Limit cannot be negative)',
                               api_tools.limit_field)

    @mock_test_field(return_value='foo')
    def test_limit_invalid_value(self, get_mock):
        self.assertRaisesRegex(utils.Error, 'Bad request',
                               api_tools.limit_field)



# Copyright 2015 NEC Corporation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import fixtures
from oslo_config import cfg

from ironic_inspector import node_cache
from ironic_inspector.pxe_filter import base as pxe_filter
from ironic_inspector.pxe_filter import iptables
from ironic_inspector.test import base as test_base
from ironic_inspector import utils

CONF = cfg.CONF


class TestIptablesDriver(test_base.NodeTest):

    def setUp(self):
        super(TestIptablesDriver, self).setUp()
        CONF.set_override('rootwrap_config', '/some/fake/path')
        # NOTE(milan) we ignore the state checking in order to avoid having to
        # always call e.g self.driver.init_filter() to set proper driver state
        self.mock_fsm = self.useFixture(
            fixtures.MockPatchObject(iptables.IptablesFilter, 'fsm')).mock
        self.mock_call = self.useFixture(
            fixtures.MockPatchObject(iptables.processutils, 'execute')).mock
        self.driver = iptables.IptablesFilter()
        self.mock_iptables = self.useFixture(
            fixtures.MockPatchObject(self.driver, '_iptables')).mock
        self.mock_should_enable_dhcp = self.useFixture(
            fixtures.MockPatchObject(iptables, '_should_enable_dhcp')).mock
        self.mock_get_inactive_macs = self.useFixture(
            fixtures.MockPatchObject(pxe_filter, 'get_inactive_macs')).mock
        self.mock_get_inactive_macs.return_value = set()
        self.mock_get_active_macs = self.useFixture(
            fixtures.MockPatchObject(pxe_filter, 'get_active_macs')).mock
        self.mock_get_active_macs.return_value = set()
        self.mock_ironic = mock.Mock()
        self.mock_ironic.ports.return_value = []

    def check_fsm(self, events):
        # assert the iptables.fsm.process_event() was called with the events
        calls = [mock.call(event) for event in events]
        self.assertEqual(calls, self.driver.fsm.process_event.call_args_list)

    def test_init_args(self):
        self.driver.init_filter()
        init_expected_args = [
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport', '67',
             '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-N', self.driver.chain)]

        call_args_list = self.mock_iptables.call_args_list

        for (args, call) in zip(init_expected_args, call_args_list):
            self.assertEqual(args, call[0])

        expected = ('sudo', 'ironic-inspector-rootwrap', CONF.rootwrap_config,
                    'iptables', '-w')
        self.assertEqual(expected, self.driver.base_command)
        self.check_fsm([pxe_filter.Events.initialize])

    def test_init_args_old_iptables(self):
        exc = iptables.processutils.ProcessExecutionError(2, '')
        self.mock_call.side_effect = exc
        self.driver.init_filter()
        init_expected_args = [
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport', '67',
             '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-N', self.driver.chain)]

        call_args_list = self.mock_iptables.call_args_list

        for (args, call) in zip(init_expected_args, call_args_list):
            self.assertEqual(args, call[0])

        expected = ('sudo', 'ironic-inspector-rootwrap', CONF.rootwrap_config,
                    'iptables',)
        self.assertEqual(expected, self.driver.base_command)
        self.check_fsm([pxe_filter.Events.initialize])

    def test_init_kwargs(self):
        self.driver.init_filter()
        init_expected_kwargs = [
            {'ignore': True},
            {'ignore': True},
            {'ignore': True}]

        call_args_list = self.mock_iptables.call_args_list

        for (kwargs, call) in zip(init_expected_kwargs, call_args_list):
            self.assertEqual(kwargs, call[1])
        self.check_fsm([pxe_filter.Events.initialize])

    def test_init_fails(self):
        class MyError(Exception):
            pass

        self.mock_call.side_effect = MyError('Oops!')
        self.assertRaisesRegex(MyError, 'Oops!', self.driver.init_filter)
        self.check_fsm([pxe_filter.Events.initialize, pxe_filter.Events.reset])

    def _test__iptables_args(self, expected_port):
        self.driver = iptables.IptablesFilter()
        self.mock_iptables = self.useFixture(
            fixtures.MockPatchObject(self.driver, '_iptables')).mock
        self.mock_should_enable_dhcp.return_value = True

        _iptables_expected_args = [
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-F', self.driver.new_chain),
            ('-X', self.driver.new_chain),
            ('-N', self.driver.new_chain),
            ('-A', self.driver.new_chain, '-j', 'ACCEPT'),
            ('-I', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-E', self.driver.new_chain, self.driver.chain)
        ]

        self.driver.sync(self.mock_ironic)
        call_args_list = self.mock_iptables.call_args_list

        for (args, call) in zip(_iptables_expected_args,
                                call_args_list):
            self.assertEqual(args, call[0])
        self.mock_get_inactive_macs.assert_called_once_with(self.mock_ironic)
        self.check_fsm([pxe_filter.Events.sync])

    def test__iptables_args_ipv4(self):
        CONF.set_override('ip_version', '4', 'iptables')
        self._test__iptables_args('67')

    def test__iptables_args_ipv6(self):
        CONF.set_override('ip_version', '6', 'iptables')
        self._test__iptables_args('547')

    def test__iptables_kwargs(self):
        _iptables_expected_kwargs = [
            {'ignore': True},
            {'ignore': True},
            {'ignore': True},
            {},
            {},
            {},
            {'ignore': True},
            {'ignore': True},
            {'ignore': True}
        ]

        self.driver.sync(self.mock_ironic)
        call_args_list = self.mock_iptables.call_args_list

        for (kwargs, call) in zip(_iptables_expected_kwargs,
                                  call_args_list):
            self.assertEqual(kwargs, call[1])
        self.check_fsm([pxe_filter.Events.sync])

    def _test_sync_with_denylist(self, expected_port):
        self.driver = iptables.IptablesFilter()
        self.mock_iptables = self.useFixture(
            fixtures.MockPatchObject(self.driver, '_iptables')).mock
        self.mock_get_active_macs.return_value = ['AA:BB:CC:DD:EE:FF']
        self.mock_get_inactive_macs.return_value = ['FF:EE:DD:CC:BB:AA']
        self.mock_should_enable_dhcp.return_value = True

        _iptables_expected_args = [
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-F', self.driver.new_chain),
            ('-X', self.driver.new_chain),
            ('-N', self.driver.new_chain),
            # deny
            ('-A', self.driver.new_chain, '-m', 'mac', '--mac-source',
             self.mock_get_inactive_macs.return_value[0], '-j', 'DROP'),
            ('-A', self.driver.new_chain, '-j', 'ACCEPT'),
            ('-I', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-E', self.driver.new_chain, self.driver.chain)
        ]

        self.driver.sync(self.mock_ironic)
        self.check_fsm([pxe_filter.Events.sync])
        call_args_list = self.mock_iptables.call_args_list

        for (args, call) in zip(_iptables_expected_args,
                                call_args_list):
            self.assertEqual(args, call[0])
        self.mock_get_inactive_macs.assert_called_once_with(self.mock_ironic)

        # check caching

        self.mock_iptables.reset_mock()
        self.mock_get_inactive_macs.reset_mock()
        self.driver.sync(self.mock_ironic)
        self.mock_get_inactive_macs.assert_called_once_with(self.mock_ironic)
        self.assertFalse(self.mock_iptables.called)

    def test_sync_with_denylist_ipv4(self):
        CONF.set_override('ip_version', '4', 'iptables')
        self._test_sync_with_denylist('67')

    def test_sync_with_denylist_ipv6(self):
        CONF.set_override('ip_version', '6', 'iptables')
        self._test_sync_with_denylist('547')

    def _test_sync_with_allowlist(self, expected_port):
        CONF.set_override('deny_unknown_macs', True, 'pxe_filter')
        self.driver = iptables.IptablesFilter()
        self.mock_iptables = self.useFixture(
            fixtures.MockPatchObject(self.driver, '_iptables')).mock
        self.mock_get_active_macs.return_value = ['AA:BB:CC:DD:EE:FF']
        self.mock_get_inactive_macs.return_value = ['FF:EE:DD:CC:BB:AA']
        self.mock_should_enable_dhcp.return_value = True

        _iptables_expected_args = [
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-F', self.driver.new_chain),
            ('-X', self.driver.new_chain),
            ('-N', self.driver.new_chain),
            # deny
            ('-A', self.driver.new_chain, '-m', 'mac', '--mac-source',
             self.mock_get_active_macs.return_value[0], '-j', 'ACCEPT'),
            ('-A', self.driver.new_chain, '-j', 'DROP'),
            ('-I', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-E', self.driver.new_chain, self.driver.chain)
        ]

        self.driver.sync(self.mock_ironic)
        self.check_fsm([pxe_filter.Events.sync])
        call_args_list = self.mock_iptables.call_args_list

        for (args, call) in zip(_iptables_expected_args,
                                call_args_list):
            self.assertEqual(args, call[0])
        self.mock_get_active_macs.assert_called_once_with(self.mock_ironic)

        # check caching

        self.mock_iptables.reset_mock()
        self.mock_get_active_macs.reset_mock()
        self.driver.sync(self.mock_ironic)
        self.mock_get_active_macs.assert_called_once_with(self.mock_ironic)
        self.assertFalse(self.mock_iptables.called)

    def test_sync_with_allowlist_ipv4(self):
        CONF.set_override('ip_version', '4', 'iptables')
        self._test_sync_with_allowlist('67')

    def test_sync_with_allowlist_ipv6(self):
        CONF.set_override('ip_version', '6', 'iptables')
        self._test_sync_with_allowlist('547')

    def _test__iptables_clean_cache_on_error(self, expected_port):
        self.driver = iptables.IptablesFilter()
        self.mock_iptables = self.useFixture(
            fixtures.MockPatchObject(self.driver, '_iptables')).mock
        self.mock_get_inactive_macs.return_value = ['AA:BB:CC:DD:EE:FF']
        self.mock_should_enable_dhcp.return_value = True

        self.mock_iptables.side_effect = [None, None, RuntimeError('Oops!'),
                                          None, None, None, None, None, None]
        self.assertRaises(RuntimeError, self.driver.sync, self.mock_ironic)
        self.check_fsm([pxe_filter.Events.sync, pxe_filter.Events.reset])
        self.mock_get_inactive_macs.assert_called_once_with(self.mock_ironic)

        # check caching
        syncs_expected_args = [
            # driver reset
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-F', self.driver.new_chain),
            ('-X', self.driver.new_chain),
            ('-N', self.driver.new_chain),
            # deny
            ('-A', self.driver.new_chain, '-m', 'mac', '--mac-source',
             self.mock_get_inactive_macs.return_value[0], '-j', 'DROP'),
            ('-A', self.driver.new_chain, '-j', 'ACCEPT'),
            ('-I', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.new_chain),
            ('-D', 'INPUT', '-i', 'br-ctlplane', '-p', 'udp', '--dport',
             expected_port, '-j', self.driver.chain),
            ('-F', self.driver.chain),
            ('-X', self.driver.chain),
            ('-E', self.driver.new_chain, self.driver.chain)
        ]

        self.mock_iptables.reset_mock()
        self.mock_iptables.side_effect = None
        self.mock_get_inactive_macs.reset_mock()
        self.mock_fsm.reset_mock()
        self.driver.sync(self.mock_ironic)
        self.check_fsm([pxe_filter.Events.sync])
        call_args_list = self.mock_iptables.call_args_list

        for (idx, (args, call)) in enumerate(zip(syncs_expected_args,
                                                 call_args_list)):
            self.assertEqual(args, call[0], 'idx: %s' % idx)
        self.mock_get_inactive_macs.assert_called_once_with(self.mock_ironic)

    def test__iptables_clean_cache_on_error_ipv4(self):
        CONF.set_override('ip_version', '4', 'iptables')
        self._test__iptables_clean_cache_on_error('67')

    def test__iptables_clean_cache_on_error_ipv6(self):
        CONF.set_override('ip_version', '6', 'iptables')
        self._test__iptables_clean_cache_on_error('547')

    def test_iptables_command_ipv4(self):
        CONF.set_override('ip_version', '4', 'iptables')
        driver = iptables.IptablesFilter()
        self.assertEqual(driver._cmd_iptables, 'iptables')

    def test_iptables_command_ipv6(self):
        CONF.set_override('ip_version', '6', 'iptables')
        driver = iptables.IptablesFilter()
        self.assertEqual(driver._cmd_iptables, 'ip6tables')

    def test_deny_unknown_macs_and_node_not_found_hook_bad(self):
        CONF.set_override('deny_unknown_macs', True, 'pxe_filter')
        CONF.set_override('node_not_found_hook', True, 'processing')
        self.assertRaisesRegex(utils.Error, 'Configuration error:',
                               self.driver.__init__)


class Test_ShouldEnableDhcp(test_base.BaseTest):
    def setUp(self):
        super(Test_ShouldEnableDhcp, self).setUp()
        self.mock_introspection_active = self.useFixture(
            fixtures.MockPatchObject(node_cache, 'introspection_active')).mock

    def test_introspection_active(self):
        self.mock_introspection_active.return_value = True
        self.assertIs(True, iptables._should_enable_dhcp())

    def test_node_not_found_hook_set(self):
        # DHCP should be always opened if node_not_found hook is set
        CONF.set_override('node_not_found_hook', 'enroll', 'processing')
        self.mock_introspection_active.return_value = False
        self.assertIs(True, iptables._should_enable_dhcp())

    def test__should_enable_dhcp_false(self):
        self.mock_introspection_active.return_value = False
        self.assertIs(False, iptables._should_enable_dhcp())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic_inspector.common.i18n import _


DEFAULT_CPU_FLAGS_MAPPING = {
    'vmx': 'cpu_vt',
    'svm': 'cpu_vt',
    'aes': 'cpu_aes',
    'pse': 'cpu_hugepages',
    'pdpe1gb': 'cpu_hugepages_1g',
    'smx': 'cpu_txt',
}


_OPTS = [
    cfg.BoolOpt('boot_mode',
                default=False,
                help=_('Whether to store the boot mode (BIOS or UEFI).')),
    cfg.DictOpt('cpu_flags',
                default=DEFAULT_CPU_FLAGS_MAPPING,
                help=_('Mapping between a CPU flag and a capability to set '
                       'if this flag is present.')),
]


def register_opts(conf):
    conf.register_opts(_OPTS, 'capabilities')


def list_opts():
    return _OPTS



# -*- coding: utf-8 -*-
#

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinxcontrib.apidoc',
              'sphinx.ext.viewcode',
              'sphinxcontrib.rsvgconverter',
              'oslo_policy.sphinxext',
              'oslo_policy.sphinxpolicygen',
              'oslo_config.sphinxext',
              'oslo_config.sphinxconfiggen']

try:
    import openstackdocstheme
    extensions.append('openstackdocstheme')
except ImportError:
    openstackdocstheme = None

openstackdocs_repo_name = 'openstack/ironic-inspector'
openstackdocs_pdf_link = True
openstackdocs_use_storyboard = True
openstackdocs_projects = [
    'bifrost',
    'devstack',
    'ironic',
    'ironic-python-agent',
    'oslo.rootwrap',
    'python-ironicclient',
    'python-ironic-inspector-client',
    'tooz',
]

wsme_protocols = ['restjson']

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = u'OpenStack Developers'

config_generator_config_file = '../../tools/config-generator.conf'
sample_config_basename = '_static/ironic-inspector'

policy_generator_config_file = '../../tools/policy-generator.conf'
sample_policy_basename = '_static/ironic-inspector'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
#from ironic import version as ironic_version
# The full version, including alpha/beta/rc tags.
#release = ironic_version.version_info.release_string()
# The short X.Y version.
#version = ironic_version.version_info.version_string()

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['ironic.']

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# NOTE(cinerama): mock out nova modules so docs can build without warnings
#import mock
#import sys
#MOCK_MODULES = ['nova', 'nova.compute', 'nova.context']
#for module in MOCK_MODULES:
#    sys.modules[module] = mock.Mock()

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
if openstackdocstheme is not None:
    html_theme = 'openstackdocs'
else:
    html_theme = 'default'
#html_theme_path = ["."]
#html_theme = '_theme'
#html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'ironic-inspectordoc'

latex_use_xindy = False

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        'index',
        'doc-ironic-inspector.tex',
        u'Ironic Inspector Documentation',
        u'OpenStack Foundation',
        'manual'
    ),
]

# -- Options for seqdiag ------------------------------------------------------

seqdiag_html_image_format = "SVG"

# -- sphinxcontrib.apidoc configuration --------------------------------------

apidoc_module_dir = '../../ironic_inspector'
apidoc_output_dir = 'contributor/api'
apidoc_excluded_paths = [
    'migrations',
    'test',
    'common/i18n*'
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

from oslo_config import cfg

from ironic_inspector.common.i18n import _


_OPTS = [
    cfg.StrOpt('listen_address',
               default='::',
               help=_('IP to listen on.')),
    cfg.PortOpt('listen_port',
                default=5050,
                help=_('Port to listen on.')),
    cfg.StrOpt('host',
               default=socket.getfqdn(),
               sample_default='localhost',
               help=_('Name of this node. This can be an opaque identifier. '
                      'It is not necessarily a hostname, FQDN, or IP address. '
                      'However, the node name must be valid within '
                      'an AMQP key, and if using ZeroMQ, a valid '
                      'hostname, FQDN, or IP address.')),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               choices=[('noauth', _('no authentication')),
                        ('keystone', _('use the Identity service for '
                                       'authentication')),
                        ('http_basic', _('HTTP basic authentication'))],
               help=_('Authentication method used on the ironic-inspector '
                      'API. "noauth", "keystone" or "http_basic" are valid '
                      'options. "noauth" will disable all authentication.')),
    cfg.StrOpt('http_basic_auth_user_file',
               default='/etc/ironic-inspector/htpasswd',
               help=_('Path to Apache format user authentication file used '
                      'when auth_strategy=http_basic')),
    cfg.IntOpt('timeout',
               default=3600,
               # We're using timedelta which can overflow if somebody sets this
               # too high, so limit to a sane value of 10 years.
               max=315576000,
               help=_('Timeout after which introspection is considered '
                      'failed, set to 0 to disable.')),
    cfg.IntOpt('clean_up_period',
               default=60,
               min=0,
               help=_('Amount of time in seconds, after which repeat clean up '
                      'of timed out nodes and old nodes status information. '
                      'WARNING: If set to a value of 0, then the periodic '
                      'task is disabled and inspector will not sync with '
                      'ironic to complete the internal clean-up process. '
                      'Not advisable if the deployment uses a PXE filter, '
                      'and will result in the ironic-inspector ceasing '
                      'periodic cleanup activities.')),
    cfg.IntOpt('leader_election_interval',
               default=10,
               help=_('Interval (in seconds) between leader elections.')),
    cfg.BoolOpt('use_ssl',
                default=False,
                help=_('SSL Enabled/Disabled')),
    cfg.IntOpt('max_concurrency',
               default=1000, min=2,
               help=_('The green thread pool size.')),
    cfg.IntOpt('introspection_delay',
               default=5,
               help=_('Delay (in seconds) between two introspections. Only '
                      'applies when boot is managed by ironic-inspector (i.e. '
                      'manage_boot==True).')),
    cfg.ListOpt('ipmi_address_fields',
                default=['redfish_address', 'ilo_address', 'drac_host',
                         'drac_address'],
                help=_('Ironic driver_info fields that are equivalent '
                       'to ipmi_address.')),
    cfg.StrOpt('rootwrap_config',
               default="/etc/ironic-inspector/rootwrap.conf",
               help=_('Path to the rootwrap configuration file to use for '
                      'running commands as root')),
    cfg.IntOpt('api_max_limit', default=1000, min=1,
               help=_('Limit the number of elements an API list-call '
                      'returns')),
    cfg.BoolOpt('can_manage_boot', default=True,
                help=_('Whether the current installation of ironic-inspector '
                       'can manage PXE booting of nodes. If set to False, '
                       'the API will reject introspection requests with '
                       'manage_boot missing or set to True.')),
    cfg.BoolOpt('enable_mdns', default=False,
                help=_('Whether to enable publishing the ironic-inspector API '
                       'endpoint via multicast DNS.')),
    cfg.BoolOpt('standalone', default=True,
                help=_('Whether to run ironic-inspector as a standalone '
                       'service. It\'s EXPERIMENTAL to set to False.')),
]


def register_opts(conf):
    conf.register_opts(_OPTS)


def list_opts():
    return _OPTS



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log
import tooz
from tooz import coordination

from ironic_inspector import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)

COORDINATION_PREFIX = 'ironic_inspector'
COORDINATION_GROUP_NAME = '.'.join([COORDINATION_PREFIX, 'service_group'])
LOCK_PREFIX = 'ironic_inspector.'


class Coordinator(object):
    """Tooz coordination wrapper."""

    group_name = COORDINATION_GROUP_NAME.encode('ascii')
    lock_prefix = LOCK_PREFIX

    def __init__(self, prefix=None):
        """Creates a coordinator instance for service coordination.

        :param prefix: The prefix to be part of the member id of the service.
                       Different types of services on the same host should use
                       different prefix to work properly.
        """
        self.coordinator = None
        self.started = False
        self.prefix = prefix if prefix else 'default'
        self.is_leader = False
        self.supports_election = True

    def start(self, heartbeat=True):
        """Start coordinator.

        :param heartbeat: Whether spawns a new thread to keep heartbeating with
                          the tooz backend. Unless there is periodic task to
                          do heartbeat manually, it should be always set to
                          True.
        """
        if self.started:
            return

        member_id = '.'.join([COORDINATION_PREFIX, self.prefix,
                             CONF.host]).encode('ascii')
        self.coordinator = coordination.get_coordinator(
            CONF.coordination.backend_url, member_id)
        self.coordinator.start(start_heart=heartbeat)
        self.started = True
        LOG.debug('Coordinator started successfully.')

    def stop(self):
        """Disconnect from coordination backend and stop heartbeat."""
        if self.started:
            try:
                self.coordinator.stop()
            except Exception as e:
                LOG.error('Failed to stop coordinator: %s', e)
            self.coordinator = None
            self.started = False
            LOG.debug('Coordinator stopped successfully')

    def _validate_state(self):
        if not self.started:
            raise utils.Error('Coordinator should be started before '
                              'executing coordination actions.')

    def _create_group(self):
        try:
            request = self.coordinator.create_group(self.group_name)
            request.get()
        except coordination.GroupAlreadyExist:
            LOG.debug('Group %s already exists.', self.group_name)

    def _join_election(self):
        self.is_leader = False

        def _when_elected(event):
            LOG.info('This conductor instance is a group leader now.')
            self.is_leader = True

        try:
            self.coordinator.watch_elected_as_leader(
                self.group_name, _when_elected)
            self.coordinator.run_elect_coordinator()
        except tooz.NotImplemented:
            LOG.warning('The coordination backend does not support leader '
                        'elections, assuming we are a leader. This is '
                        'deprecated, please use a supported backend.')
            self.is_leader = True
            self.supports_election = False

    def join_group(self):
        """Join service group."""
        self._validate_state()
        try:
            request = self.coordinator.join_group(self.group_name)
            request.get()
        except coordination.GroupNotCreated:
            self._create_group()
            request = self.coordinator.join_group(self.group_name)
            request.get()
        except coordination.MemberAlreadyExist:
            pass

        self._join_election()
        LOG.debug('Joined group %s', self.group_name)

    def leave_group(self):
        """Leave service group"""
        self._validate_state()
        try:
            request = self.coordinator.leave_group(self.group_name)
            request.get()
            LOG.debug('Left group %s', self.group_name)
        except coordination.MemberNotJoined:
            LOG.debug('Leaving a non-existing group.')

    def get_members(self):
        """Get members in the service group."""
        self._validate_state()
        try:
            result = self.coordinator.get_members(self.group_name)
            return result.get()
        except coordination.GroupNotCreated:
            # If the group does not exist, there should be no members in it.
            return set()

    def get_lock(self, uuid):
        """Get lock for node uuid."""
        self._validate_state()
        lock_name = (self.lock_prefix + uuid).encode('ascii')
        return self.coordinator.get_lock(lock_name)

    def run_elect_coordinator(self):
        """Trigger a new leader election."""
        if self.supports_election:
            LOG.debug('Starting leader election')
            self.coordinator.run_elect_coordinator()
            LOG.debug('Finished leader election')
        else:
            LOG.warning('The coordination backend does not support leader '
                        'elections, assuming we are a leader. This is '
                        'deprecated, please use a supported backend.')
            self.is_leader = True


_COORDINATOR = None


@lockutils.synchronized('inspector_coordinator')
def get_coordinator(prefix=None):
    global _COORDINATOR
    if _COORDINATOR is None:
        _COORDINATOR = Coordinator(prefix=prefix)
    return _COORDINATOR
